@@ -66,7 +66,7 @@ class Const(Op):
     def matchable(self, op, specials=False):
         return isinstance(op, Const)
     def matches(self, vals):
-        return ( v for v in vals if self.value == v )
+        return (v for v in vals if self.value == v)
 
 
 class Numeric(Const):
@@ -100,16 +100,16 @@ class String(Const):
 
 
 class Pattern(Op):
-    @property
-    def value(self):
-        return self.args[0]
+    def __repr__(self):
+        return str(self.value)
     def matchable(self, op, specials=False):
         raise NotImplementedError
 
 
 class Wildcard(Pattern):
-    def __repr__(self):
-        return f'*'
+    @property
+    def value(self):
+        return '*'
     def matches(self, vals):
         return iter(v for v in vals if v is not NOP)
     def matchable(self, op, specials=False):
@@ -117,8 +117,9 @@ class Wildcard(Pattern):
 
 
 class WildcardFirst(Wildcard):
-    def __repr__(self):
-        return f'*?'
+    @property
+    def value(self):
+        return '*?'
     def matches(self, vals):
         v = next(super().matches(vals), _marker)
         return iter(() if v is _marker else (v,))
@@ -129,21 +130,32 @@ class WildcardFirst(Wildcard):
 
 class Regex(Pattern):
     @property
+    def value(self):
+        return f'/{self.args[0]}/'
+    @property
     def pattern(self):
-        return re.compile(self.value)
-    def __repr__(self):
-        return f'/{self.value}/'
+        return re.compile(self.args[0])
     def matches(self, vals):
-        vals = ( v for v in vals if v is not NOP )
-        iterable = ( self.pattern.fullmatch(v) for v in vals )
-        return ( m[0] for m in iterable if m )
+        vals = (v for v in vals if v is not NOP)
+        vals = {v if isinstance(v, (str, bytes)) else str(v): v for v in vals}
+        iterable = (self.pattern.fullmatch(v) for v in vals)
+        # we want to regex match numerics as strings but return numerics
+        # unless they were transformed, of course
+        for m in iterable:
+            if not m:
+                continue
+            if m[0] != m.string:
+                yield m[0]
+            else:
+                yield vals[m.string]
     def matchable(self, op, specials=False):
         return isinstance(op, Const) or (specials and isinstance(op, (Special, Regex)))
 
 
 class RegexFirst(Regex):
-    def __repr__(self):
-        return f'/{self.value}/?'
+    @property
+    def value(self):
+        return f'/{self.args[0]}/?'
     def matches(self, vals):
         iterable = super().matches(vals)
         v = next(iterable, _marker)
@@ -159,23 +171,94 @@ class Special(Op):
     def matchable(self, op, specials=False):
         return isinstance(op, Special)
     def matches(self, vals):
-        return ( v for v in vals if v == self.value )
+        return (v for v in vals if v == self.value)
 
 
 class Appender(Special):
-    def __repr__(self):
+    @property
+    def value(self):
         return '+'
     def matchable(self, op, specials=False):
         return isinstance(op, Appender)
     def matches(self, vals):
-        return ( v for v in vals if self.value in v )
+        return (v for v in vals if self.value in v)
 
 
 class AppenderUnique(Appender):
-    def __repr__(self):
+    @property
+    def value(self):
         return '+?'
 
 
+class FilterOp(Op):
+    def is_pattern(self):
+        return False
+    def is_filtered(self, node):
+        raise NotImplementedError
+
+
+class FilterKeyValue(FilterOp):
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self.kv = tuple((k, v) for k, v in self.args)
+    def __hash__(self):
+        return hash(self.kv)
+    def __repr__(self):
+        return ','.join(f'{k}={v}' for k, v in self.kv)
+
+    def is_filtered(self, node):
+        if not hasattr(node, 'keys'):
+            return False
+        # conjunctive evaluation
+        for k, v in self.kv:
+            for km in k.matches(node.keys()):
+                for vm in v.matches((node[km],)):
+                    break
+                else:
+                    return False
+                break
+            else:
+                return False
+        return True
+
+    def match_one(self, item):
+        if not isinstance(item, FilterKeyValue):
+            return None
+        r = ()
+        for k, v in self.kv:
+            found = False
+            for ik, iv in item.kv:
+                if not k.matchable(ik) or not v.matchable(iv):
+                    continue
+                mk = next(k.matches((ik.value,)), _marker)
+                mv = next(v.matches((iv.value,)), _marker)
+                if _marker in (mk, mv):
+                    continue
+                if (mk, mv) not in r:
+                    r += ((mk, mv),)
+                found = True
+            if not found:
+                return None
+        return r
+
+    def matches(self, items):
+        r = ()
+        for item in items:
+            o = self.match_one(item)
+            if o is None:
+                return
+            if not o:
+                continue
+            r += (type(self)(*o),)
+        yield r
+
+
+def is_filtered(node, filters):
+    for f in filters:
+        if not f.is_filtered(node):
+            return False
+    return True
+
 
 #
 #
@@ -184,15 +267,24 @@ def itemof(node, val):
     return val if isinstance(node, (str, bytes)) else node.__class__([val])
 
 
-class Empty(Op):
+
+class CmdOp(Op):
+    pass
+
+
+class Empty(CmdOp):
     def __repr__(self):
+        if self.args:
+            return '.'.join(repr(a) for a in self.args)
         return ''
     def is_pattern(self):
         return False
     def operator(self, top=False):
         return ''
     def values(self, node):
-        return (node,)
+        if is_filtered(node, self.args):
+            return (node,)
+        return ()
     def default(self):
         return ''
     def match(self, op, specials=False):
@@ -201,47 +293,88 @@ class Empty(Op):
         return None
 
 
-class Key(Op):
+class Key(CmdOp):
     @classmethod
     def concrete(cls, val):
         import numbers
         if isinstance(val, numbers.Number):
             return cls(NumericQuoted(val))
         return cls(Word(val))
-    @property
-    def op(self):
-        return self.args[0]
+
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self.op = self.args[0]
+        self.filters = self.args[1:]
+
     def is_pattern(self):
         return isinstance(self.op, Pattern)
+
     def __repr__(self):
-        return f'{self.op}'
+        return '.'.join(repr(a) for a in self.args)
+
     def operator(self, top=False):
         s = quote(self.op.value)
         return s if top else '.' + s
-    def keys(self, node):
-        if not hasattr(node, 'keys'):
-            return ()
-        return self.op.matches(node.keys())
+
+    def is_filtered(self, node):
+        return is_filtered(node, self.filters)
+
     def items(self, node):
-        for k in self.keys(node):
+        if not hasattr(node, 'keys'):
+            return
+        keys = self.op.matches(node.keys()) if self.op is not None else node.keys()
+        for k in keys:
             try:
-                yield k, node[k]
-            except TypeError:
-                pass
+                v = node[k]
+            except (KeyError, TypeError, IndexError):
+                continue
+            if self.is_filtered(v):
+                yield (k, v)
+
+    def keys(self, node):
+        return (k for k, _ in self.items(node))
     def values(self, node):
-        return ( v for _,v in self.items(node) )
+        return (v for _, v in self.items(node))
+
     def is_empty(self, node):
         return not tuple(self.keys(node))
+
     def default(self):
         if self.is_pattern():
             return {}
-        return {self.op.value: None}
+        if not self.filters:
+            return {self.op.value: None}
+        return {self.op.value: {}}
 
     def match(self, op, specials=False):
-        if not self.op.matchable(op.op, specials):
+        if not isinstance(op, Key):
             return None
-        val = next(self.op.matches((op.op.value,)), _marker)
-        return None if val is _marker else Match(val)
+        results = ()
+        if self.op is None: # always matchable since None is implicit wilcard
+            if op.op is not None:
+                results += (Match(op.op.value),)
+        elif op.op is None:
+            if not isinstance(self.op, Wildcard):
+                return None
+        elif not self.op.matchable(op.op, specials):
+            return None
+        else:
+            # match key
+            val = next(self.op.matches((op.op.value,)), _marker)
+            if val is _marker:
+                return None
+            results += (Match(val),)
+
+        # match filters
+        found = ()
+        for f in self.filters:
+            m = next(f.matches(op.filters), _marker)
+            if m is _marker:
+                return None
+            found += m
+        results += tuple(Match(f) for f in found)
+
+        return results
 
     def update(self, node, key, val):
         node[key] = self.default() if val is ANY else val
@@ -270,29 +403,44 @@ class Slot(Key):
         if isinstance(val, numbers.Number):
             return cls(Numeric(val))
         return String(val)
+
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        if isinstance(self.op, FilterOp):
+            self.filters = self.args
+            self.op = None
     def __repr__(self):
-        return f'[{self.op}]'
+        return '[' + super().__repr__()  + ']'
+
+    def is_pattern(self):
+        # a pure filter treats op as an implicit wildcard
+        return super().is_pattern() or bool(self.filters)
+
     def operator(self, top=False):
-        return '[' + quote(self.op.value, as_key=False) + ']'
-        s = str(self.op)
-        if isinstance(s, Word):
-            try:
-                int(s)
-                if isinstance(self.op, str):
-                    s = repr(s)
-            except ValueError:
-                pass
-        return '[' + s + ']'
-    def keys(self, node):
+        iterable = (repr(a) for a in self.filters)
+        if self.op is not None:
+            iterable = itertools.chain((quote(self.op.value, as_key=False),), iterable)
+        return '[' + '.'.join(iterable) + ']'
+
+    def items(self, node):
         if hasattr(node, 'keys'):
-            return super().keys(node)
-        if self.is_pattern():
-            return self.op.matches(idx for idx,_ in enumerate(node))
-        try:
-            _ = node[self.op.value]
-            return (self.op.value,)
-        except (KeyError, IndexError):
-            return ()
+            yield from super().items(node)
+            return
+
+        if self.op is None:
+            keys = (idx for idx, _ in enumerate(node))
+        elif self.is_pattern():
+            keys = self.op.matches(idx for idx, _ in enumerate(node))
+        else:
+            keys = (self.op.value,)
+        for k in keys:
+            try:
+                v = node[k]
+            except (KeyError, TypeError, IndexError):
+                continue
+            if self.is_filtered(v):
+                yield (k, v)
+
     def default(self):
         if isinstance(self.op, Numeric) and self.op.is_int():
             return []
@@ -311,7 +459,11 @@ class Slot(Key):
         except TypeError:
             node = node[:key] + itemof(node, val) + node[key+1:]
         return node
+
     def upsert(self, node, val):
+        if hasattr(node, 'keys'):
+            if self.op is None:
+                raise RuntimeError('Bracket updates not permitted on dicts with naked op')
         return super().upsert(node, val)
 
     def pop(self, node, key):
@@ -353,16 +505,13 @@ class SlotSpecial(Slot):
         return cls(val)
     def default(self):
         return []
-    def keys(self, node):
-        return (-1,)
+
     def items(self, node):
-        for k in self.keys(node):
-            try:
-                yield k, node[k]
-            except TypeError:
-                pass
-    def values(self, node):
-        return ( v for _,v in self.items(node) )
+        try:
+            yield -1, node[-1]
+        except TypeError:
+            pass
+
     def is_empty(self, node):
         return True
 
@@ -383,7 +532,7 @@ class SlotSpecial(Slot):
         return node
 
 
-class Slice(Op):
+class Slice(CmdOp):
     @classmethod
     def concrete(cls, val):
         o = cls()
@@ -439,12 +588,17 @@ class Slice(Op):
         if stop == '+':
             return 1 << 64
         return max(0, int((stop - start) / step))
+
     def keys(self, node):
         return (self.slice(node),)
     def items(self, node):
-        return ( (k,node[k]) for k in self.keys(node) )
+        for k in self.keys(node):
+            try:
+                yield (k, node[k])
+            except (TypeError, KeyError, IndexError):
+                pass
     def values(self, node):
-        return ( v for _,v in self.items(node) )
+        return (v for _, v in self.items(node))
     def is_empty(self, node):
         return not node[self.slice(node)]
     def default(self):
@@ -499,7 +653,7 @@ class Slice(Op):
         return node
 
 
-class Invert(Op):
+class Invert(CmdOp):
     @classmethod
     def concrete(cls, val):
         return cls(val)
@@ -530,6 +684,7 @@ class rdoc(str):
         title = 'Supported transforms\n\n'
         return title + '\n'.join(f'{name}\t{fn.__doc__ or ""}' for name,fn in Dotted._registry.items())
 
+
 class Dotted:
     _registry = {}
 
@@ -543,6 +698,7 @@ class Dotted:
     def __init__(self, results):
         self.ops = tuple(results['ops'])
         self.transforms = tuple(tuple(r) for r in results.get('transforms', ()))
+
     def assemble(self, start=0):
         return assemble(self, start)
     def __repr__(self):
@@ -580,6 +736,8 @@ def quote(key, as_key=True):
             s = f"#'{key}'"
         else:
             s = str(key)
+    elif isinstance(key, Op):
+        return str(key)
     else:
         raise NotImplementedError
     return s
@@ -646,7 +804,7 @@ def updates(ops, node, val, has_defaults=False):
     if cur.is_empty(node) and not has_defaults:
         built = updates(ops, build_default(ops), val, True)
         return cur.upsert(node, built)
-    for k,v in cur.items(node):
+    for k, v in cur.items(node):
         node = cur.update(node, k, updates(ops, v, val, has_defaults))
     return node
 
